
from flask.testing import FlaskClient
from flask import json
from werkzeug import cached_property, urls


class JsonResponseMixin(object):
    '''Mixin which adds method to jsonify response data'''
    @cached_property
    def json(self):
        return json.loads(self.data)


def make_client_response(response_class):
    class ClientResponse(response_class, JsonResponseMixin):
        pass

    return ClientResponse


class Client(FlaskClient):
    def get(self, url, params=None, **kargs):
        if params:
            url = urls.Href(url)(params)

        return super(Client, self).get(url, **kargs)

    def post(self, url, data, **kargs):
        kargs['data'] = data
        return super(Client, self).post(url, **kargs)

    def put(self, url, data, **kargs):
        kargs['data'] = data
        return super(Client, self).put(url, **kargs)

    def patch(self, url, data, **kargs):
        kargs['data'] = data
        return super(Client, self).patch(url, **kargs)


class JSONClient(Client):
    def open(self, *args, **kargs):
        # all requests will be treated like JSON unless otherwise specified
        kargs.setdefault('content_type', 'application/json')

        # convert kargs['data'] dict to json, otherwise, werkzeug will cast to ImmutableDict
        # ImmutableDicts are a problem since our API performs updates to incoming data
        # in normal operation, all dict data is sent as JSON strings which our API then converts to a dict which can be updated
        if isinstance(kargs.get('data'), dict):
            try:
                kargs['data'] = json.dumps(kargs['data'])
            except Exception:
                # some data may not be serializable (e.g. StringIO for file uploads)
                # in that case, it's probably ok to not serialize since that data dict won't be updated in the API
                pass

        return super(Client, self).open(*args, **kargs)




from collections import defaultdict

import core

def create_app(package_name, config=None, options=None, **flask_options):
    if options is None:
        options = {}

    app = core.FlaskCarafe(package_name, **flask_options)

    app.config.from_object(config)

    # use default dict for convenience since each key represents keyword args to init_app() functions
    opts = defaultdict(dict)
    opts.update(options)

    if app.config.get('CARAFE_JSON_ENABLED', True):
        core.json.init_app(app, **opts['json'])

    if app.config.get('CARAFE_SESSION_ENABLED', True):
        core.session.init_app(app, **opts['session'])

    if app.config.get('CARAFE_AUTH_ENABLED', True):
        core.auth.init_app(app, **opts['auth'])

    if app.config.get('CARAFE_CACHE_ENABLED', True):
        core.cache.init_app(app, **opts['cache'])

    if app.config.get('CARAFE_LOGGER_ENABLED', True):
        core.logger.init_app(app, **opts['cache'])

    return app




from .app import FlaskCarafe

from .ext.json import jsonify, JSONEncoder
from .ext.logger import Logger
from .ext.cache import Cache
from .ext.signaler import signaler
from .ext.auth import Auth

# extensions for use
# each object below should expose an "init_app" function/method
from .ext import json
from .ext import session
logger = Logger()
cache = Cache(signaler=signaler)
auth = Auth()



from flask.sessions import SecureCookieSessionInterface

class SessionInterface(SecureCookieSessionInterface):
    def __init__(self, salt=None, permanent=True):
        if salt:
            # salt is added on top of secret key
            self.salt = salt

        self.permanent = permanent

    def open_session(self, app, request):
        session = super(SessionInterface, self).open_session(app, request)

        if session is not None and self.permanent:
            # set session to permanent
            session.permanent = True

        return session

def init_app(app):
	app.config.setdefault('CARAFE_SESSION_PERMANENT', True)
	app.config.setdefault('CARAFE_SESSION_SALT', None)

	permanent = app.config['CARAFE_SESSION_PERMANENT'] and app.config['PERMANENT_SESSION_LIFETIME'].total_seconds() > 0

	app.session_interface = SessionInterface(salt=app.config['CARAFE_SESSION_SALT'], permanent=permanent)




from functools import wraps

from flask import json, jsonify as _jsonify
from werkzeug.exceptions import HTTPException, default_exceptions

class JSONEncoder(json.JSONEncoder):
    '''Extend flask's JSONEncoder class'''
    def default(self, o):
        if hasattr(o, 'isoformat'):
            return o.isoformat()
        else:
            return super(JSONEncoder, self).default(o)

def json_error_handler(ex):
    '''Convert exception to jsonify response'''
    error = {
        'description': getattr(ex, 'description', ''),
        'message': str(ex),
        'name': getattr(ex, 'name', ''),
        'code': ex.code if isinstance(ex, HTTPException) else 500
    }

    for k, v in error.iteritems():
        # only json-able values should be sent
        try:
            json.dumps(v)
        except Exception:
            error[k] = ''

    response = jsonify(error=error)
    response.status_code = error['code']

    return response

def jsonify(f=None, *args, **kargs):
    '''Function or decorator that returns jsonfiy response'''
    if callable(f):
        @wraps(f)
        def decorated(*args, **kargs):
            return _jsonify(**f(*args, **kargs))
        return decorated
    else:
        if f is not None:
            # consider `f` a positional arg
            args = tuple([f] + list(args))
        return _jsonify(*args, **kargs)

def init_app(app, error_handler=json_error_handler, encoder=JSONEncoder):
    if encoder:
        app.json_encoder = encoder

    if error_handler:
        for code in default_exceptions.iterkeys():
            app.error_handler_spec[None][code] = error_handler



from functools import wraps, partial
from threading import Thread
import re

class classproperty(object):
    '''
    Decorator that adds class properties.
    Allows for usage like @property but applies the property at the class level.
    Helps avoid having to use metaclasses or other complex techniques to achieve similar results.
    '''
    def __init__(self, getter):
        self.getter = getter
    def __get__(self, instance, owner):
        return self.getter(owner)


def iterflatten(l):
    '''
    Return iterator which flattens list/tuple of lists/tuples
    >>> assert list(iterflatten([1, [2,3], [4, [5, [6]], 7], 8])) == [1,2,3,4,5,6,7,8]
    '''
    for x in l:
        if isinstance(x, (list, tuple)):
            for y in flatten(x):
                yield y
        else:
            yield x

def flatten(l):
    '''
    Return flattened list of a list/tuple of lists/tuples
    >>> assert flatten([1, [2,3], [4, [5, [6]], 7], 8]) == [1,2,3,4,5,6,7,8]
    '''
    return list(iterflatten(l))

def async(f):
    '''
    Poor man's decorator to execute a function asynchronously using separate thread.
    This is an extremely limited async method which doesn't provide any way to
    access the created thread. Use with caution!

    For more robust async, consider using alternative methods/libraries (e.g. gevent)
    '''
    @wraps(f)
    def wrapper(*args, **kargs):
        thr = Thread(target=f, args=args, kwargs=kargs)
        thr.start()
    return wrapper

def to_dict(data=None, namespace=None):
    '''
    Decorator enabled version of `_to_dict()`
    '''
    __to_dict = partial(_to_dict, namespace=namespace)

    if data is None:
        def decorator(f):
            @wraps(f)
            def decorated(*args, **kargs):
                data = f(*args, **kargs)
                return __to_dict(data)
            return decorated
        return decorator
    else:
        return __to_dict(data)

def _to_dict(data, namespace=None):
    '''
    Converts elements of `data` using `data.to_dict()` or `data[].to_dict()`.
    Optionally namespaces data into a container dict using `namespace` as key.
    '''
    if isinstance(data, list):
        data = [d.to_dict() if hasattr(d, 'to_dict') else d for d in data]
    elif hasattr(data, 'to_dict'):
        data = data.to_dict()

    if namespace is not None:
        data = {namespace: data}

    return data

def urlpathjoin(*paths):
    '''Join URL paths into single URL while maintaining leading and trailing slashes
    if present on first and last elements respectively.

    >>> assert urlpathjoin('') == ''
    >>> assert urlpathjoin(['', '/a']) == '/a'
    >>> assert urlpathjoin(['a', '/']) == 'a/'
    >>> assert urlpathjoin(['', '/a', '', '', 'b']) == '/a/b'
    >>> assert urlpathjoin(['/a/', 'b/', '/c', 'd', 'e/']) == '/a/b/c/d/e/'
    >>> assert urlpathjoin(['a', 'b', 'c']) == 'a/b/c'
    >>> assert urlpathjoin(['a/b', '/c/d/', '/e/f']) == 'a/b/c/d/e/f'
    >>> assert urlpathjoin('/', 'a', 'b', 'c', 1, '/') == '/a/b/c/1/'
    >>> assert urlpathjoin([]) == ''
    '''
    paths = [str(path) for path in flatten(paths) if path]
    leading = '/' if paths and paths[0].startswith('/') else ''
    trailing = '/' if paths and paths[-1].endswith('/') else ''
    url = leading + '/'.join([p.strip('/') for p in paths if p.strip('/')]) + trailing
    return url

def camelcase_to_underscore(name):
    '''
    >>> assert camelcase_to_underscore('FooBar') == 'foo_bar'
    >>> assert camelcase_to_underscore('FooBar_Baz') == 'foo_bar__baz'
    '''
    first_cap_re = re.compile('(.)([A-Z][a-z]+)')
    all_cap_re = re.compile('([a-z0-9])([A-Z])')

    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()





import logging
from logging import getLogger, Formatter
from logging.handlers import SMTPHandler as SMTPHandlerBase, RotatingFileHandler as RotatingFileHandlerBase
import pprint
from functools import partial

from flask import request, current_app

# universal error message format for custom logger handlers
# any handler who uses this should use the HandlerDataMixin class so that extra data is attached
ERROR_FORMAT = '''
================================================================================
Message type:       %(levelname)s
Location:           %(pathname)s:%(lineno)d
Module:             %(module)s
Function:           %(funcName)s
Time:               %(asctime)s


REQUEST:

%(environ)s


ERROR:

%(message)s
================================================================================
'''

WARNING_FORMAT = '%(asctime)s: %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'

# default pprint.pformat function to use to output request environ
pformat = partial(pprint.pformat, indent=4, depth=None)

class Logger(object):
    def __init__(self, app=None):
        self.app = app

        if app: # pragma: no cover
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault('CARAFE_LOGGER_ENABLED', True)

        if app.debug or not app.config['CARAFE_LOGGER_ENABLED']: # pragma: no cover
            return

        if app.config.get('CARAFE_LOGGER_RFILE_ENABLED'):
            loggers = [app.logger] + app.config.get('CARAFE_LOGGER_RFILE_ADD_LOGGERS', [])
            self.add_handlers(loggers, create_rotating_file_handler(app.config))

        if app.config.get('CARAFE_LOGGER_SMTP_ENABLED'):
            loggers = [app.logger] + app.config.get('CARAFE_LOGGER_SMTP_ADD_LOGGERS', [])
            self.add_handlers(loggers, create_email_handler(app.config))

    def add_handlers(self, loggers, handlers):
        if not isinstance(loggers, list): # pragma: no cover
            loggers = [loggers]

        if not isinstance(handlers, list): # pragma: no cover
            handlers = [handlers]

        for logger in loggers:
            if isinstance(logger, basestring):
                logger = getLogger(logger)

            for handler in handlers:
                logger.addHandler(handler)

    def __getattr__(self, attr):
        # proxy attribute calls to current_app.logger as shortcut
        return getattr(current_app.logger, attr)

class HandlerDataMixin(object):
    def update_record(self, record):
        environ = request.environ.copy()
        environ.update({
            'REQUEST_DATA': request.data,
            'REQUEST_JSON': request.get_json(silent=True),
            'REQUEST_FORM': request.form.to_dict(),
            'REQUEST_FILES': request.files.to_dict(),
            'REQUEST_HEADERS': dict(request.headers),
            'REQUEST_ARGS': request.args.to_dict()
        })

        record.environ = pformat(environ)

    def emit(self, record):
        self.update_record(record)
        super(HandlerDataMixin, self).emit(record)

class SMTPHandler(HandlerDataMixin, SMTPHandlerBase):
    '''Override `emit()` function so that request data is available to log formatter'''

    def getSubject(self, record):
        '''Override with custom subject'''
        return '[{0}] Application Error: {1}'.format(current_app.name, request.url)

class RotatingFileHandler(HandlerDataMixin, RotatingFileHandlerBase):
    '''Override `emit()` function so that request data is available to log formatter'''
    pass

def create_rotating_file_handler(options):
    kargs = {
        'filename': options['CARAFE_LOGGER_RFILE_FILENAME'],
        'mode': options.get('CARAFE_LOGGER_RFILE_MODE', 'a'),
        'maxBytes': options.get('CARAFE_LOGGER_RFILE_MAXBYTES', 0),
        'backupCount': options.get('CARAFE_LOGGER_RFILE_BACKUPCOUNT', 0),
        'encoding': options.get('CARAFE_LOGGER_RFILE_ENCODING'),
        'delay': options.get('CARAFE_LOGGER_RFILE_DELAY', 0)
    }

    handler = RotatingFileHandler(**kargs)
    handler.setLevel(getattr(logging, options.get('CARAFE_LOGGER_RFILE_LEVEL', 'WARNING')))

    handler.setFormatter(Formatter(WARNING_FORMAT))

    return handler

def create_email_handler(options):
    if isinstance(options['CARAFE_LOGGER_SMTP_TOADDRS'], basestring):
        # convert to list
        options['CARAFE_LOGGER_SMTP_TOADDRS'] = map(lambda x: x.strip(), options['CARAFE_LOGGER_SMTP_TOADDRS'].split(','))

    kargs = {
        'mailhost': (options['CARAFE_LOGGER_SMTP_SERVER'], options['CARAFE_LOGGER_SMTP_PORT']),
        'fromaddr': options['CARAFE_LOGGER_SMTP_FROMADDR'],
        'toaddrs': options['CARAFE_LOGGER_SMTP_TOADDRS'],
        'subject': 'Application Error',
        'credentials': (options['CARAFE_LOGGER_SMTP_USERNAME'], options['CARAFE_LOGGER_SMTP_PASSWORD']),
        'secure': () if options.get('CARAFE_LOGGER_SMTP_USE_TLS') else None
    }

    handler = SMTPHandler(**kargs)
    handler.setLevel(getattr(logging, options.get('CARAFE_LOGGER_SMTP_LEVEL', 'ERROR')))

    handler.setFormatter(Formatter(ERROR_FORMAT))

    return handler




from flask.signals import Namespace

class Signaler(object):
    '''
    Collection of supported signals to use throughout application

    Define new signals as properties and return self.namespace.signal('name')
    '''
    def __init__(self, namespace=None):
        if not namespace:
            namespace = Namespace()

        self.namespace = namespace
        self._signals = {}

    def send(self, signal, *args, **kargs):
        '''Send signal by name'''
        getattr(self, signal).send(*args, **kargs)

    def connect(self, signal, *args, **kargs):
        '''Connect to signl by name'''
        getattr(self, signal).connect(*args, **kargs)

    def make_signal(self, signal):
        if signal not in self._signals:
            self._signals[signal] = self.namespace.signal(signal)

        return self._signals[signal]

    def __getattr__(self, signal):
        '''Return signal object. If it doesn't exist in registry, create it.'''
        return self.make_signal(signal)

signaler = Signaler()




from flask import Response as ResponseBase, json, current_app, request

class Response(ResponseBase):
    def __init__(self, content=None, *args, **kargs):
        if isinstance(content, (list, dict)):
            kargs['mimetype'] = 'application/json'
            content = to_json(content)

        super(Response, self).__init__(content, *args, **kargs)

def to_json(content):
    indent = None
    if current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] and not request.is_xhr:
        indent = 2
    return json.dumps(content, indent=indent)





from functools import wraps

from flask import request, abort, current_app
from flask.ext.classy import FlaskView, _FlaskViewMeta, route

from .signaler import signaler
from ..utils import urlpathjoin, _to_dict, camelcase_to_underscore


def register_view(app, classes, route_prefix='', route_base='', subdomain=None, trailing_slash=False, replace_prefix=False, replace_base=False):
    '''
    Register a group of FlaskView based classes.

    If `replace_prefix` or `replace_base` is True, then
    class route_<...> is replaced instead of extended.
    '''

    if not isinstance(classes, list):
        classes = [classes]

    _prefix = []
    if route_prefix:
        _prefix.append(route_prefix)

    _base = []
    if route_base:
        _base.append(route_base)

    for cls in classes:
        prefix = _prefix[:]
        base = _base[:]

        if cls.route_prefix is not None and not replace_prefix:
            prefix.append(cls.route_prefix)

        if cls.route_base is not None and not replace_base:
            base.append(cls.route_base)

        cls.register(
            app,
            route_prefix=urlpathjoin(prefix),
            route_base=urlpathjoin(base),
            subdomain=subdomain,
            trailing_slash=trailing_slash
        )

def to_dict(func):
    '''Wrapper around `self._to_dict` which supports namespacing as defined at the class level.'''
    @wraps(func)
    def decorated(*args, **kargs):
        data = func(*args, **kargs)

        if not isinstance(data, current_app.response_class):
            # only convert to dict if function return is not Response

            # if decorator called from FlaskView.decorators[],
            # `self` won't be passed in so we need to get it from `func`
            if hasattr(func, 'im_self'):
                self = func.im_self
            else:
                self = args[0]

            data = self._to_dict(data)

        return data
    return decorated


def noop(*args, **kargs):
    pass


class MetaView(_FlaskViewMeta):
    def __new__(cls, name, bases, dct):
        if 'permissions' in dct:
            # if permissions dict defined on class,
            # then set corresponding permissions keys on class dict
            # and decorate with supplied functions
            for method, permissions in dct['permissions'].iteritems():
                dct[method] = dct.get(method, noop)

                if not isinstance(permissions, list):
                    permissions = [permissions]

                for permission in permissions:
                    dct[method] = permission(dct[method])

        return type.__new__(cls, name, bases, dct)


class BaseView(FlaskView):
    __metaclass__ = MetaView

    # route prefix
    route_prefix = None

    # set route_base to '' to disable FlaskView's auto-routing from class name
    route_base = ''

    # controller class which has hooks for index(), get(), post(), put(), patch(), and delete()
    __controller__ = None

    # namespace to apply to returned data when returning json response
    dict_namespace = None

    # set this to override automatic cache namespace generation from class name
    cache_namespace = None

    # when invalidating this class' cache keys, also invalidate cache keys with these namespaces
    cache_cascade = []

    # universal decorators applied to all view routes
    decorators = []

    # assign permissions to instance methods
    # `keys` correspond to method that has permission requirements
    # `values` can be a single permission function or a list of functions which will wrap/decorate method
    # e.g.
    #   permissions = {
    #       'before_post': auth.require.admin(401),
    #       'delete': [auth.require.admin(403), auth.require.auth(401)]
    #   }
    # @note: be aware that permissions are merged when subclassing
    # so if a parent class defines a partial set of permissions,
    # the subclass will have all the parent permissions as well as its own.
    # however, if a subclass overwrites a parent permission, the subclass
    # permission will be the only one used.
    # essentially, subclasses is like calling `Parent.permissions.update(Subclass.permissions)`
    permissions = {}

    @property
    def request(self):
        '''Proxy to request object'''
        return request

    @property
    def _dict_namespace(self):
        namespace = getattr(self, 'dict_namespace', None)

        if namespace is True:
            class_name = self.__class__.__name__
            if class_name.lower().endswith('view'):
                class_name = class_name[:-4]
            namespace = camelcase_to_underscore(class_name)

        return namespace

    @property
    def controller(self):
        # override as needed for custom initialization
        return self.__controller__()

    def after_request(self, name, response):
        # send `after_<http-method>` signal
        signaler.send('after_' + request.method.lower(), self)

        return response

    def _to_dict(self, data):
        return _to_dict(data, namespace=self._dict_namespace)

    def _results_or_404(self, results):
        if results is None:
            abort(404)
        return results

    def _index(self):
        return self.controller.index(params=self.request.args)

    def _get(self, _id):
        results = self.controller.get(_id=_id)
        return self._results_or_404(results)

    def _post(self):
        return self.controller.post(self.request.data)

    def _put(self, _id):
        results = self.controller.put(_id, self.request.data)
        return self._results_or_404(results)

    def _patch(self, _id):
        ctrl = self.controller
        results = getattr(ctrl, 'patch', ctrl.put)(_id, self.request.data)
        return self._results_or_404(results)

    def _delete(self, _id):
        results = self.controller.delete(_id)
        return self._results_or_404(results)


# @note: Each of the Rest endpoints below comes with a corresponding `_<method>` function.
# The purpose of this is to make it easier to subclass a method without then having to also
# duplicate the decoratores applied to the public method. Also, in the event that caching
# is enabled on an endpoint and another endpoint needs a non-cached version, or if `super()`
# needs to be called, then the `_<method>` function can be used to bypass the cache.

class RestView(BaseView):
    '''REST read/write endpoints'''

    decorators = [to_dict]

    def index(self):
        return self._index()

    def get(self, _id):
        return self._get(_id)

    def post(self):
        return self._post()

    def put(self, _id):
        return self._put(_id)

    def patch(self, _id):
        return self._patch(_id)

    def delete(self, _id):
        return self._delete(_id)




import factory
import core
import ext
import utils

from .client import Client, JSONClient
from .factory import create_app
from .core import (
	jsonify,
	json,
	session,
	logger,
	signaler,
	cache,
	auth
)

from .__meta__ import (
    __title__,
    __summary__,
    __url__,
    __version__,
    __author__,
    __email__,
    __license__
)




from flask import Flask, request
from flask._compat import text_type, string_types

from .request import Request
from .response import Response

class FlaskCarafe(Flask):
    request_class = Request
    response_class = Response

    def make_response(self, rv):
        status_or_headers = headers = None
        if isinstance(rv, tuple):
            rv, status_or_headers, headers = rv + (None,) * (3 - len(rv))

        if rv is None:
            raise ValueError('View function did not return a response')

        if isinstance(status_or_headers, (dict, list)):
            headers, status_or_headers = status_or_headers, None

        if not isinstance(rv, self.response_class):
            # add support for converting list or dict to json response
            if isinstance(rv, (text_type, bytes, bytearray, list, dict)):
                rv = self.response_class(rv, headers=headers, status=status_or_headers)
                headers = status_or_headers = None
            else:
                rv = self.response_class.force_type(rv, request.environ)

        if status_or_headers is not None:
            if isinstance(status_or_headers, string_types):
                rv.status = status_or_headers
            else:
                rv.status_code = status_or_headers

        if headers:
            rv.headers.extend(headers)

        return rv




from functools import wraps
import inspect

from flask import request, current_app
from werkzeug import urls
from flask.ext.cache import Cache as CacheBase


class Cache(CacheBase):
    '''
    Manager class for handling creating and deleting cache keys based on API view events
    '''

    view_key_format = '{namespace}:view:{path}'

    def __init__(self, app=None, with_jinja2_ext=True, config=None, signaler=None):
        self.signaler = signaler
        super(Cache, self).__init__(app=app, with_jinja2_ext=with_jinja2_ext, config=config)

    def init_app(self, app, config=None):
        self.app = app
        if config is None:
            config = self.config

        if config is None:
            config = app.config

        config.setdefault('CARAFE_CACHE_ENABLED', True)
        config.setdefault('CARAFE_CACHE_IGNORED_REQUEST_ARGS', [])

        if config['CARAFE_CACHE_ENABLED']:
            super(Cache, self).init_app(app, config=config)

            if self.signaler:
                self.connect_signals()

    def connect_signals(self):
        # @note: have to keep a handle on the signal for the receive event to work (not sure why exactly)
        # not ok: self.signaler.<signal>.connect(self.<method>)
        # ok: self.<signal_name> = self.signaler.<signal>
        #     self.<signal_name>.connect(self.<method>)

        # connect receivers
        self.signaler.after_post.connect(self.after_post)
        self.signaler.after_put.connect(self.after_put)
        self.signaler.after_patch.connect(self.after_patch)
        self.signaler.after_delete.connect(self.after_delete)

    def get_cache_namespace(self, obj):
        if getattr(obj, 'cache_namespace', None) is not None:
            return obj.cache_namespace
        elif hasattr(obj, '__name__'):
            return obj.__name__
        else:
            return obj.__class__.__name__

    @property
    def client(self):
        '''Proxy to cache client wrapper.'''
        return self.cache if self.enabled else None

    @property
    def server(self):
        '''Proxy to cache server client.'''
        return getattr(self.cache, '_client', None) if self.enabled else None

    @property
    def enabled(self):
        return current_app.config['CARAFE_CACHE_ENABLED']

    def clear_keys(self, pipe, *keys, **kargs):
        '''Clear specified keys'''
        if not keys:
            return

        execute = kargs.get('execute')

        keys = [current_app.config['CACHE_KEY_PREFIX'] + k for k in keys]
        pipe.delete(*keys)

        if execute:
            pipe.execute()

        return True

    def clear_prefix(self, pipe, *prefixes, **kargs):
        '''Clear keys starting with prefix'''
        keys = []
        for prefix in prefixes:
            keys.extend(pipe.keys(prefix + '*'))

        return self.clear_keys(pipe, *keys, **kargs)

    def clear(self, prefixes=None, keys=None):
        '''
        Clear cache keys using an optional prefix, regex, and/or list of keys

        :param string prefix: prefix to append to global CACHE_KEY_PREFIX
        :param string regex: regex to filter keys matched by CACHE_KEY_PREFIX + prefix
        :param list keys: additional keys to delete during pipeline transaction
        '''
        if not self.enabled:
            return

        if not any([prefixes, keys]):
            # this is the same as clearing the entire cache
            return self.cache.clear()

        if not hasattr(self.server, 'pipeline'):
            # enhanced cache clearing is only supported using redis
            return self.cache.clear()

        if prefixes is None:
            prefixes = []

        if keys is None:
            keys = []

        with self.server.pipeline() as pipe:
            self.clear_prefix(pipe, *prefixes)
            self.clear_keys(pipe, *keys)
            pipe.execute()

        return True

    def cached_view(self, timeout=None, namespace=None, unless=None, include_request_args=True):
        '''
        Wrapper around self.cached which itself is a decorator.
        We're wrapping because we want to have access to the class instance of the view in order to namespace the key.
        We can't always namespace using key_prefix since some cache decorators are placed around parent classes which
        don't know anything about the child class.
        '''

        def wrap(f):
            @wraps(f)
            def wrapper(*args, **kargs):
                if not self.enabled:
                    return f(*args, **kargs)

                if namespace is not None:
                    # make namespace available in case `f` is used as signal sender
                    # mainly used to get namespace when invalidating cache keys via namespace prefix
                    f.cache_namespace = namespace

                # if args[0] is set, then this is a class based view, else use function
                obj = args[0] if args else f
                cache_namespace = self.get_cache_namespace(obj)
                view_path = self.create_view_path(include_request_args)
                key_prefix = self.view_key_format.format(namespace=cache_namespace, path=view_path, **request.view_args)

                cached = self.cached(timeout=timeout, key_prefix=key_prefix, unless=unless)(f)

                try:
                    # cache server could be down
                    result = cached(*args, **kargs)
                except Exception as e: # pragma: no cover
                    # return function call instead
                    result = f(*args, **kargs)
                    current_app.logger.exception(e)

                return result

            return wrapper

        return wrap

    def create_view_path(self, include_request_args=False):
        '''Construct view path from request.path with option to include GET args'''
        href = urls.Href(request.path)

        if include_request_args:
            args = {k:v for k, v in request.args.iteritems() if k not in current_app.config['CARAFE_CACHE_IGNORED_REQUEST_ARGS']}
        else:
            args = None

        return href(args)

    def on_modified_record(self, sender):
        '''Common tasks to perform when a record is modified'''
        namespace = self.get_cache_namespace(sender)
        prefixes = [self.view_key_format.format(namespace=namespace, path='')]

        # append cascade keys which should be invalidated (typically due to this API's records being used in other APIs)
        prefixes += getattr(sender, 'cache_cascade', [])

        try:
            self.clear(prefixes=prefixes)
        except Exception as e: # pragma: no cover
            current_app.logger.exception(e)

    def after_post(self, sender):
        '''Handle the `after_post` event. Executed after a ModelAPI POST request.'''
        self.on_modified_record(sender)

    def after_put(self, sender):
        '''Handle the `after_put` event. Executed after a ModelAPI PUT request.'''
        self.on_modified_record(sender)

    def after_patch(self, sender):
        '''Handle the `after_patch` event. Executed after a ModelAPI PATCH request.'''
        self.on_modified_record(sender)

    def after_delete(self, sender):
        '''Handle the `after_delete` event. Executed after a ModelAPI DELETE request.'''
        self.on_modified_record(sender)

